from typing import List
from os.path import isdir, isfile

import requests

from plantit_cli.store.store import Store
from plantit_cli.store.util import list_files
from plantit_cli.config import Config


class Terrain(Store):

    @property
    def run(self):
        return self.__run

    def __init__(self, run: Config):
        self.__run = run

    def list_directory(self) -> List[str]:
        """
        Lists files in the collection.

        Returns:
            A list of all files in the collection.
        """

        response = requests.get(
            f"https://de.cyverse.org/terrain/secured/filesystem/paged-directory?limit=1000&path={self.__run.input['from']}",
            headers={'Authorization': f"Bearer {self.__run.cyverse_token}"})
        if response.status_code == 401:
            raise RuntimeError('CyVerse authentication cyverse_token expired or invalid')
        content = response.json()
        files = [file['path'] for file in content['files']]
        return files

    def download_directory(self, to_path, patterns=None):
        """
        Pulls all files in the collection matching a given pattern to the local path.

        Args:
            to_path: The local path.
            patterns: The file pattern.
        """

        from_paths = [p for p in self.list() if patterns in p] if patterns is not None else self.list()
        print(f"Preparing to pull {len(from_paths)} files")
        for from_path in from_paths:
            full_to_path = f"{to_path}/{from_path.split('/')[-1]}"
            print(f"Pulling '{from_path}' to '{full_to_path}'")
            with requests.get(f"https://de.cyverse.org/terrain/secured/fileio/download?path={from_path}",
                              headers={'Authorization': f"Bearer {self.__run.cyverse_token}"}) as response:
                if response.status_code == 401:
                    raise RuntimeError('CyVerse authentication cyverse_token expired or invalid')
                # with open(f"{from_path}/{from_path.split('/')[-1]}", 'wb') as file:
                with open(full_to_path, 'wb') as file:
                    for chunk in response.iter_content(chunk_size=8192):
                        # If you have chunk encoded response uncomment if
                        # and set chunk_size parameter to None.
                        # if chunk:
                        file.write(chunk)

    def upload_directory(self, from_path, include_pattern=None):
        """
        Pushes all files matching a given pattern from the local path to the collection.

        Args:
            from_path: The local path.
            include_pattern: The file pattern.
        """

        is_local_file = isfile(from_path)
        is_local_dir = isdir(from_path)
        if not (is_local_dir or is_local_file):
            raise FileNotFoundError(f"Local path '{from_path}' does not exist")
        elif is_local_dir:
            from_paths = [p for p in list_files(from_path) if include_pattern in p] if include_pattern is not None else list_files(from_path)
            print(f"Preparing to push {len(from_paths)} files")
            for from_path in [str(p) for p in from_paths]:
                print(f"Pushing '{from_path}' to '{self.__run.output['to']}'")
                response = requests.post(f"https://de.cyverse.org/terrain/secured/fileio/upload?dest={self.__run.output['to']}",
                                         headers={'Authorization': f"Bearer {self.__run.cyverse_token}"},
                                         files={'file': open(from_path, 'rb')})
                if response.status_code == 401:
                    raise RuntimeError('CyVerse authentication cyverse_token expired or invalid')
                if response.status_code == 500:
                    raise RuntimeError(f'Internal server error')
        elif is_local_file:
            print(f"Pushing {from_path} to {self.__run.output['to']}")
            response = requests.post(f"https://de.cyverse.org/terrain/secured/fileio/upload?dest={self.__run.output['to']}",
                                     headers={'Authorization': f"Bearer {self.__run.cyverse_token}"},
                                     files={'file': open(from_path, 'rb')})
            if response.status_code == 401:
                raise RuntimeError('CyVerse authentication cyverse_token expired or invalid')
            if response.status_code == 500:
                raise RuntimeError(f'Internal server error')
        else:
            raise ValueError(
                f"Cannot overwrite object '{self.__run.output['to']}' with contents of local directory '{from_path}' (specify a remote directory instead)")
